#!/usr/bin/perl
# Author: Alex Efros <powerman@consultant.com>
# Desc -- Process template wroten in "iCGI" style - break usual CGI to 2 
#     pieces of code: "logic" and "interface".
# Idea -- take "interface" code from HTML-template, process it in 
#     eperl-style and output it.
#
package POWER::iCGI;
use CGI;
use List::Util qw( min );
use POWER::Utils;
use strict;
our ($VERSION, $ERR2BROWSER, %HEADER, %COOKIE, $skip_stack, $path, $__, $PKG);
$VERSION    = '2.73';
$ERR2BROWSER= 0;
$PKG        = __PACKAGE__;
# don't output in stack trace: mod_perl's internal handlers
# FIXME magical value 4 may be incorrect in another mod_perl versions
$skip_stack = ($ENV{GATEWAY_INTERFACE}||'') =~ /Perl/i ? 4 : 0; # -w happy
sub import {
    no strict;
    *{"${\scalar caller}::$_"} = \&{$_} for qw(required show redirect_to
        SendFile);
    *{"${\scalar caller}::$_"} = \%{$_} for qw(HEADER COOKIE);
}
sub _cleanup {
    # under mod_perl $SIG{__DIE__} executed after _print by unknown reason
    undef $SIG{__DIE__} if $SIG{__DIE__} == \&DIE;
    # for mod_perl: clean up global variables
    (%HEADER, %COOKIE) = ();
}
# print all parameters and exit
sub _print {
    _cleanup();
    $|=1;
    print @_;
    exit()
}
# print `die' message to browser (with file name and stack trace)
# and to STDERR (with PID)
# FIXME probably this should not be here, POWER::Utils is much more suitable
# place for general-purpose $SIG{__DIE__} handler
sub DIE {
    my $msg = "iCGI ERROR ($path):\n$_[0]"; chomp $msg; $msg .= "\n\n";
    my $i; $msg .= (caller $i)[3]."(...) at ".(caller $i)[1].
        " line ".(caller $i)[2]."\n" while caller ++$i+$skip_stack;
    warn "($$) ".$msg."($$) END OF iCGI STACK TRACE\n";
    _print($ERR2BROWSER ? CGI::header("text/plain").$msg : "");
}
# just like CGI::redirect but with %COOKIE and %HEADER support
sub redirect_to { 
    _print(CGI::redirect(-uri=>$_[0], -cookie=>[values %COOKIE], %HEADER));
}
# don't output in stack trace: show() OR required(), _Eval(), Eval(), do()
sub _Eval {
    local $skip_stack = $skip_stack + 4;
    Eval($_[0]);
}
# force newline after every bit of perl code and restore previous line number;
# needed in this case:   <!--& #comment -->@~code()~@
# _restoreline(1, "filename.pl") -- initialization
# _restoreline("some\ntext\n")   -- calculate and return line number
{ my ($line, $filename);
sub _restoreline {
    if (@_ == 2) { ($line, $filename) = @_ } else { $line += $_[0] =~ tr/\n// }
    return("\n#line $line \"$filename\"\n");
}}
# process and execute iCGI-style HTML template
sub show {
    local $SIG{__DIE__} = \&DIE unless $SIG{__DIE__};
    my $e    = "package ${\scalar caller}; no strict; local (\$^W) = 0 ;";
    $_[0]=~m!^/! ? ($path=$_[0]) : ($path =~ s{[^/]*$}{$_[0]});
    $e .= _restoreline(1, ($path =~ m/([^\/]+)$/)[0]);
    my $s = eval { Cat($path) };
    die "Can't open '$path': $@" if $@;
    while ( 1 ) {
        ($s=~/\G<!--&(.*?)-->/gcs)  && do {
            $e .= "$1;";
            $e .= _restoreline($1);
            }
     || ($s=~/\G@~(.*?)~@/gcs)      && do {
            $e .= "\$${PKG}::__ .= POWER::Utils::Enc(''.(do { $1; })) ;";
            $e .= _restoreline($1);
            }
     || ($s=~/\G#~(.*?)~#/gcs)      && do {
            $e .= "\$${PKG}::__ .= do { $1; } ;";
            $e .= _restoreline($1);
            }
     || ($s=~/\G\^~(.*?)~\^/gcs)    && do {
            $e .= "\$${PKG}::__ .= POWER::Utils::EncUri(''.(do { $1; })) ;";
            $e .= _restoreline($1);
            }
     || ($s=~/\G(.*?)(?=<!--&|@~|#~|\^~|\n|$)(\n?)/gcsm) && do {
            $e .= "\$${PKG}::__ .= \"\Q$1\E" . ($2 ? '\n' : '') . "\";$2";
            _restoreline($1.$2);
            }
     || last;
    }
    # use $__ as accumulator for calculated HTML result
    undef $__; _Eval($e); die $@ if $@;
    _print(CGI::header(-cookie => [values %COOKIE], %HEADER), $__);
}
# for mod_perl: reload require'd modules if these modules changed
my %inc; # closure :-)
sub required {
    local $SIG{__DIE__} = \&DIE unless $SIG{__DIE__};
    my $e="package ${\scalar caller}; no strict; local (\$^W) = 0 ;\n\n";
    for (@_) {
        next if exists $inc{$_} and $inc{$_} == (stat)[9];
        # FIXME    local ..  $path  regex.
        # FIXME   $path s///     ''  .
        local $path = ''; # change $path only for processing current file
        m!^/! ? ($path=$_) : ($path=~s{[^/]*$}{$_});
        my $e = $e . _restoreline(1, $path);
        die "Can't read file: `$path'" unless -f $path and -r $path and
            defined(my $content = Cat($path));
        _Eval $e . $content; die $@ if $@;
        $inc{$_} = (stat)[9];
    }
}
# send file or scalar to browser
# BUG Opera don't understand Content-Length by unknown reason...
sub SendFile {
    my $file = shift;
    my %p = (type=>"application/x-download", range=>not(ref $file),
        cache=>0, inline=>0, @_);
    open my $F, "<", $file or die "Can't open file: $!";
    my $len = ref $file ? length(${$file}) : -s $file;
    my ($start, $end) = $p{range}
        ? ($ENV{HTTP_RANGE}=~/^bytes=(\d*)-(\d*)$/i) : ();
    $end   = $len-1 if $end eq "";
    $end   = $len-1 if ($end  +=0) < 0 or $end   >= $len;
    $start = 0      if ($start+=0) < 0 or $start >  $end;
    my $left = $end-$start+1;
    my $headers =
        "Content-Type: $p{type}\r\n".
        "Content-Disposition: ".($p{inline}?"inline":"attachment")."\r\n".
        "Content-Length: $left\r\n".
        (! $p{cache} ? "Pragma: no-cache\r\nExpires: access\r\n" : "").
        ($p{range}   ? "Accept-Range: bytes\r\n" : "").
        ($len        ? "Content-Range: bytes $start-$end/$len\r\n" : "").
        "\r\n";
    my $data = q{};
    seek($F, $start, 0);
    my ($r, $buf);
    $left-=length($buf), $data.=$buf while $r = read($F, $buf, min($left,10240));
    die "Read error: $!" unless defined $r;
    _print($headers, $data);
}
1;
